﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.JobBase;
using HeuristicLab.Hive.Slave.Common;

namespace HeuristicLab.Hive.Slave.ExecutionEngine {
  public class Executor : MarshalByRefObject, IDisposable {
    public Guid JobId { get; set; }
    public IJob Job { get; set; }
    public MessageContainer.MessageType CurrentMessage { get; set; }

    private Exception currentException;
    public String CurrentException {
      get {
        if (currentException != null) {
          return currentException.ToString();
        } else {
          return string.Empty;
        }
      }
    }
    public MessageQueue Queue { get; set; }
    
    public ExecutionState ExecutionState {
      get {
        return Job.ExecutionState;
      }
    }

    public TimeSpan ExecutionTime {
      get {
        return Job.ExecutionTime;
      }
    }

    public DateTime CreationTime { get; set; }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="serializedJob"></param>
    /// <param name="collectChildJobs">if true, all child-jobs are downloaded and the job will be resumed.</param>
    public void Start(byte[] serializedJob) {
      try {
        CreationTime = DateTime.Now;
        Job = SerializedJob.Deserialize<IJob>(serializedJob);

        RegisterJobEvents();

        if (Job.CollectChildJobs) {
          Queue.AddMessage(new MessageContainerWithCallback<SerializedJobList>(MessageContainer.MessageType.GetChildJobs, JobId,
            (childjobs) => {
              Job.Resume(childjobs.Select(j => SerializedJob.Deserialize<IJob>(j.SerializedJobData)));
            }
          ));
        } else {
          Job.Prepare();
          Job.Start();
        }
      }
      catch (OutOfMemoryException e) {
        this.currentException = e;
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.JobFailed, JobId));
      }
    }

    public void StartOnlyJob() {
      try {
        Job.Start();
      }
      catch (OutOfMemoryException e) {
        this.currentException = e;
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.JobFailed, JobId));
      }
    }

    public void Abort() {
      CurrentMessage = MessageContainer.MessageType.AbortJob;
      if ((ExecutionState == ExecutionState.Started) || (ExecutionState == ExecutionState.Paused)) {
        Job.Stop();
      } else {
        Job_JobStopped(this, EventArgs.Empty);
      }
    }

    private void RegisterJobEvents() {
      Job.JobStopped += new EventHandler(Job_JobStopped);
      Job.JobFailed += new EventHandler(Job_JobFailed);
      Job.NewChildJob += new EventHandler<EventArgs<IJob>>(Job_NewChildJob);
      Job.WaitForChildJobs += new EventHandler(Job_WaitForChildJobs);
      Job.DeleteChildJobs += new EventHandler(Job_DeleteChildJobs);
    }

    private void DeregisterJobEvents() {
      Job.JobStopped -= new EventHandler(Job_JobStopped);
      Job.JobFailed -= new EventHandler(Job_JobFailed);
      Job.NewChildJob -= new EventHandler<EventArgs<IJob>>(Job_NewChildJob);
      Job.WaitForChildJobs -= new EventHandler(Job_WaitForChildJobs);
      Job.DeleteChildJobs -= new EventHandler(Job_DeleteChildJobs);
    }

    private void Job_NewChildJob(object sender, EventArgs<IJob> e) {
      byte[] jobByteArray = SerializedJob.Serialize(e.Value);

      SerializedJob serializedJob = new SerializedJob() {
        JobInfo = new JobDto() {
          State = JobState.Offline,
          CoresNeeded = 1,
          MemoryNeeded = 0,
          PluginsNeeded = HivePluginInfoDto.FindPluginsNeeded(e.Value.GetType()),
        },
        SerializedJobData = jobByteArray
      };

      Queue.AddMessage(new MessageContainerWithJob(MessageContainer.MessageType.AddChildJob, this.JobId, serializedJob));
    }

    private void Job_WaitForChildJobs(object sender, EventArgs e) {
      // Pause the job and send it back to the hive. The server will awake it when all child-jobs are finished
      this.Job.CollectChildJobs = true;
      byte[] jobByteArray = SerializedJob.Serialize(Job);

      SerializedJob serializedJob = new SerializedJob() {
        JobInfo = new JobDto() {
          Id = JobId,
          State = JobState.WaitForChildJobs
        },
        SerializedJobData = jobByteArray
      };

      Queue.AddMessage(new MessageContainerWithJob(MessageContainer.MessageType.PauseJob, this.JobId, serializedJob));
    }

    private void Job_DeleteChildJobs(object sender, EventArgs e) {
      Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.DeleteChildJobs, this.JobId));
    }

    private void Job_JobFailed(object sender, EventArgs e) {
      HeuristicLab.Common.EventArgs<Exception> ex = (HeuristicLab.Common.EventArgs<Exception>)e;
      currentException = ex.Value;
      Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.FinishedJob, JobId));
    }

    private void Job_JobStopped(object sender, EventArgs e) {
      if (CurrentMessage == MessageContainer.MessageType.RequestSnapshot) {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.SnapshotReady, JobId));
      } else if (CurrentMessage == MessageContainer.MessageType.AbortJob) {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.JobAborted, JobId));
      } else {
        Queue.AddMessage(new MessageContainer(MessageContainer.MessageType.FinishedJob, JobId));
      }
    }

    public byte[] GetSnapshot() {
      //if the job is still running, something went VERY bad.
      if (Job.ExecutionState == Core.ExecutionState.Started) {
        return null;
      } else {
        // Clear the Status message
        CurrentMessage = MessageContainer.MessageType.NoMessage;
        // Pack the whole job inside an xml document
        byte[] job = SerializedJob.Serialize(Job);
        // Restart the job
        // Return the Snapshot
        return job;
      }
    }

    public byte[] GetFinishedJob() {
      //Job isn't finished!
      if (Job.ExecutionState == Core.ExecutionState.Started) {
        throw new InvalidStateException("Job is still running");
      } else {
        byte[] jobArr = SerializedJob.Serialize(Job);
        return jobArr;
      }
    }

    public void RequestSnapshot() {
      CurrentMessage = MessageContainer.MessageType.RequestSnapshot;
      Job.Stop();
    }

    private void RestoreJobObject(byte[] sjob) {
      Job = SerializedJob.Deserialize<IJob>(sjob);
    }

    public Executor() {
      CurrentMessage = MessageContainer.MessageType.NoMessage;
    }

    public void Dispose() {
      DeregisterJobEvents();
      Queue = null;
      Job = null;
    }
  }
}
